MODULE HLR3;

(* 
  Created 93-05-18 by Marcos C. Carrard.
  Based on H3.pas by J. Stolfi.
  Last edited by stolfi *)

IMPORT LR3, LR3Extras, LR4, LR4Extras, LR6, LR4x4, Math;

PROCEDURE FromCartesian(READONLY c: LR3.T): Point =
  BEGIN
    RETURN Point{c := LR4.T{1.0D0, c[0], c[1], c[2]}}
  END FromCartesian;
    
PROCEDURE ToCartesian(READONLY p: Point): LR3.T =
  BEGIN
    WITH w = p.c[0] DO
      <* ASSERT w # 0.0D0 *>
      RETURN LR3.T{p.c[1]/w, p.c[2]/w, p.c[3]/w}
    END
  END ToCartesian;

PROCEDURE Side(READONLY p: Point; READONLY Q: Plane): Sign =
  BEGIN
    WITH s = LR4.Dot(p.c, Q.f) DO
      IF s > 0.0D0 THEN 
        RETURN 1
      ELSIF s < 0.0D0 THEN 
        RETURN -1
      ELSE
        RETURN 0
      END
    END
  END Side;       

PROCEDURE Orient(READONLY p, q, r, s: Point): Sign =
  BEGIN
    WITH s = LR4Extras.Det(p.c, q.c, r.c, s.c) DO
      IF s > 0.0D0 THEN 
        RETURN 1
      ELSIF s < 0.0D0 THEN 
        RETURN -1
      ELSE
        RETURN 0
      END
    END
  END Orient;
  
PROCEDURE LineFromTwoPoints(READONLY p, q: Point): Line =
  BEGIN
    WITH
      a0 = p.c[0],
      a1 = p.c[1],
      a2 = p.c[2],
      a3 = p.c[3],
                 
      b0 = q.c[0],
      b1 = q.c[1],
      b2 = q.c[2],
      b3 = q.c[3]
    DO
      RETURN Line{
        k := LR6.T{
          a0*b1 - a1*b0,
          a0*b2 - a2*b0,
          a1*b2 - a2*b1,
          a0*b3 - a3*b0,
          a1*b3 - a3*b1,
          a2*b3 - a3*b2
        }
      }
    END
  END LineFromTwoPoints;
  
PROCEDURE PlaneFromThreePoints(READONLY p, q, r: Point): Plane =
  BEGIN
    RETURN Plane{f := LR4Extras.Cross(p.c, q.c, r.c)}
  END PlaneFromThreePoints;   

PROCEDURE PlaneFromLineAndPoint(READONLY n: Line; READONLY r: Point): Plane =
  BEGIN
    WITH
      a01 = n.k[0],
      a02 = n.k[1],
      a12 = n.k[2],
      a03 = n.k[3],
      a13 = n.k[4],
      a23 = n.k[5],
      
      b0 = r.c[0],
      b1 = r.c[1],
      b2 = r.c[2],
      b3 = r.c[3]
    DO
      RETURN Plane{
        f := LR4.T{
          - a12*b3 + a13*b2 - a23*b1,
          + a02*b3 - a03*b2 + a23*b0,
          - a01*b3 + a03*b1 - a13*b0,
          + a01*b2 - a02*b1 + a12*b0
        }
      }
    END
  END PlaneFromLineAndPoint;

PROCEDURE LineFromTwoPlanes(READONLY P, Q: Plane): Line =
  BEGIN
    WITH
      a0 = P.f[0],
      a1 = P.f[1],
      a2 = P.f[2],
      a3 = P.f[3],
                 
      b0 = Q.f[0],
      b1 = Q.f[1],
      b2 = Q.f[2],
      b3 = Q.f[3]
    DO
      RETURN Line{
        k := LR6.T{
          + a2*b3 - a3*b2,
          - a1*b3 + a3*b1,
          + a0*b3 - a3*b0,
          + a1*b2 - a2*b1,
          - a0*b2 + a2*b0,
          + a0*b1 - a1*b0
        }
      }
    END
  END LineFromTwoPlanes;
  
PROCEDURE PointFromThreePlanes(READONLY P, Q, R: Plane): Point =
  BEGIN
    RETURN Point{c := LR4Extras.Cross(Q.f, P.f, R.f)}
  END PointFromThreePlanes; 
  
PROCEDURE PointFromLineAndPlane(READONLY n: Line; R: Plane): Point =
  BEGIN
    WITH
      a01 = n.k[0],
      a02 = n.k[1],
      a12 = n.k[2],
      a03 = n.k[3],
      a13 = n.k[4],
      a23 = n.k[5],
      
      b123 = R.f[0],
      b023 = R.f[1],
      b013 = R.f[2],
      b012 = R.f[3]
    DO
      RETURN Point{
        c := LR4.T{
          + a01*b023 + a02*b013 + a03*b012,
          - a01*b123 + a12*b013 + a13*b012,
          - a02*b123 - a12*b023 + a23*b012,
          - a03*b123 - a13*b023 - a23*b013
        }
      }
    END
  END PointFromLineAndPlane;

PROCEDURE MapPoint(READONLY p: Point; READONLY m: PMap): Point =
  BEGIN
    RETURN Point{c := LR4x4.MapRow(p.c, m.dir)}
  END MapPoint;
        
PROCEDURE InvMapPoint(READONLY p: Point; READONLY m: PMap): Point =
  BEGIN
    RETURN Point{c := LR4x4.MapRow(p.c, m.inv)}
  END InvMapPoint;
        
PROCEDURE MapPlane(READONLY P: Plane; READONLY m: PMap): Plane =
  BEGIN
    RETURN Plane{f := LR4x4.MapCol(m.inv, P.f)}
  END MapPlane;
    
PROCEDURE InvMapPlane(READONLY P: Plane; READONLY m: PMap): Plane =
  BEGIN
    RETURN Plane{f := LR4x4.MapCol(m.dir, P.f)}
  END InvMapPlane;
    
PROCEDURE CompMap(READONLY m, n: PMap): PMap =
  BEGIN
    RETURN PMap{
      dir := LR4x4.Mul(m.dir, n.dir),
      inv := LR4x4.Mul(n.inv, m.inv)
    }
  END CompMap;    
  
PROCEDURE InvMap(READONLY m: PMap): PMap =
  BEGIN
    RETURN PMap{dir := m.inv, inv := m.dir}
  END InvMap;

PROCEDURE Normal(READONLY P: Plane): LR3.T =
  BEGIN
    WITH 
      nx = P.f[1],
      ny = P.f[2],
      nz = P.f[3],
      length = Math.hypot(Math.hypot(nx, ny), nz)
    DO
      RETURN LR3.T{nx/length, ny/length, nz/length}
    END;
  END Normal;    

PROCEDURE Dir(READONLY frm, tto: Point): LR3.T =
  BEGIN
    WITH 
      fw = frm.c[0],
      tw = tto.c[0],
                   
      fx = frm.c[1],
      tx = tto.c[1],
      dx = fw * tx - tw * fx,
      
      fy = frm.c[2],
      ty = tto.c[2],
      dy = fw * ty - tw * fy,
      
      fz = frm.c[3],
      tz = tto.c[3],
      dz = fw * tz - tw * fz,
      
      length = Math.hypot(Math.hypot(dx, dy), dz)
    DO
      RETURN LR3.T{dx/length, dy/length, dz/length}
    END;
  END Dir;  
  
PROCEDURE Dist(READONLY a, b: Point): LONGREAL =
  BEGIN
    WITH
      aw = 1.0d0/a.c[0],
      bw = 1.0d0/b.c[0],
      
      ax = a.c[1],
      bx = b.c[1],
      dx = ax*aw - bx*bw,
      
      ay = a.c[2],
      by = b.c[2],
      dy = ay*aw - by*bw,
      
      az = a.c[3],
      bz = b.c[3],
      dz = az*aw - bz*bw
    DO
      RETURN Math.hypot(Math.hypot(dx, dy), dz)
    END
  END Dist;
    
PROCEDURE DistSqr(READONLY a, b: Point): LONGREAL =
  BEGIN
    WITH 
      aw = 1.0d0/a.c[0],
      bw = 1.0d0/b.c[0],
      
      ax = a.c[1],
      bx = b.c[1],
      dx = ax*aw - bx*bw,
      
      ay = a.c[2],
      by = b.c[2],
      dy = ay*aw - by*bw,
      
      az = a.c[3],
      bz = b.c[3],
      dz = az*aw - bz*bw
    DO
      RETURN dx * dx + dy * dy + dz * dz
    END;
  END DistSqr;

PROCEDURE PerspMap(READONLY obs, foc, upp: Point): PMap =
  VAR m: PMap;
      r, s, t: LR3.T;
      d, uno: LONGREAL;
      mt, mr, mc, mrt: LR4x4.T;
  BEGIN
    <* ASSERT foc.c[0] > 0.0D0 *>  (* Focus must be finite and hither *)
    <* ASSERT obs.c[0] >= 0.0D0 *> (* Observer must be hither or infinite *)
    <* ASSERT upp.c[0] >= 0.0D0 *> (* Zenith must be hither or infinite *)

    (* Compute translation matrix *)
    FOR i := 0 TO 3 DO
      FOR j := 0 TO 3 DO
        IF i = j THEN mt[i,j] := foc.c[0];
        ELSIF i = 0 THEN mt[i,j] := -foc.c[j];
        ELSE mt[i,j] := 0.0D0;
        END;
      END;
    END;

    (* Compute rotation matrix *)
    t := Dir(foc, obs);
    s := LR3.Orthize(Dir(foc, upp), t);
    (* Zenith reference point (upp) must not be on imagesys Z axis: *)
    <* ASSERT NOT LR3.NormSqr(s) < 1.0d-12 *> 
    s := LR3.Dir(s);
    r := LR3.Dir(LR3Extras.Cross(s, t));
    
    mr[0,0] := 1.0D0;
    FOR i := 1 TO 3 DO
        mr[0,i] := 0.0D0;
        mr[i,0] := 0.0D0;
        mr[i,1] := r[i-1];
        mr[i,2] := s[i-1];
        mr[i,3] := t[i-1];
    END;

    (* Compose the two matrices: *)
    mrt := LR4x4.Mul(mt, mr);

    IF obs.c[0] = 0.0D0 THEN
      (* Observer is at infinity; cilindrical projection. *)
      m.dir := mrt;
    ELSE
      (* Observer is finite; add conical projection step. *)
      d := LR3.Dist(ToCartesian(obs), ToCartesian(foc));
      IF ABS(d) > 1.0D0 THEN
        uno := -1.0D0/d;
        d := 1.0D0;
      ELSE
        uno := -1.0D0;
      END;
      FOR i := 0 TO 3 DO
        FOR j := 0 TO 3 DO
          IF i = j THEN mc[i,j] := d ELSE mc[i,j] := 0.0D0 END;
        END;
      END;
      mc[3,0] := uno;
      m.dir := LR4x4.Mul(mrt, mc);
    END;
    
    (* Compute inverse matrix: *)
    m.inv := LR4x4.Inv(m.dir);
    RETURN m;
  END PerspMap;

BEGIN 
END HLR3.
