MODULE GaussRandom;

IMPORT Random, Math, LongReal AS LR;

VAR (*CONST*) 
  MinR: LONGREAL := Math.sqrt(Math.log(LR.MaxFinite))/LR.MaxFinite;

PROCEDURE Real(rnd: Random.T): REAL =
  BEGIN
    RETURN FLOAT(LongReal(rnd), REAL)
  END Real;

PROCEDURE LongReal(rnd: Random.T): LONGREAL =
  (* Polar method [Knuth II:3.4.1, Algorithm P] *)
  CONST SqrtTwo = 1.4142135623730950488016887242d0;
  BEGIN
    LOOP
      WITH
        u = rnd.longreal(-1.0d0, +1.0d0),
        v = rnd.longreal(-1.0d0, +1.0d0),
        r = Math.hypot(u, v)
      DO
        IF r < 1.0d0  (* In unit circle .. *)
        AND r > MinR  (* ... and far enough from overflow. *)
        THEN
          RETURN SqrtTwo * (u + v) * Math.sqrt(-Math.log(r))/r
        END
      END
    END
  END LongReal;

BEGIN
END GaussRandom.
